from manim import *

from manim_ml.neural_network.layers.embedding import EmbeddingLayer
from manim_ml.neural_network.layers.feed_forward import FeedForwardLayer
from manim_ml.neural_network.neural_network import NeuralNetwork

config.pixel_height = 720
config.pixel_width = 1280
config.frame_height = 5.0
config.frame_width = 5.0


class EmbeddingNNScene(Scene):
    def construct(self):
        embedding_layer = EmbeddingLayer()

        neural_network = NeuralNetwork(
            [
                FeedForwardLayer(5),
                FeedForwardLayer(3),
                embedding_layer,
                FeedForwardLayer(3),
                FeedForwardLayer(5),
            ]
        )

        self.play(Create(neural_network))

        self.play(neural_network.make_forward_pass_animation(run_time=5))


class TripletEmbeddingNNScene(Scene):
    def construct(self):
        embedding_layer = EmbeddingLayer()

        neural_network = NeuralNetwork(
            [
                FeedForwardLayer(5),
                FeedForwardLayer(3),
                embedding_layer,
                FeedForwardLayer(3),
                FeedForwardLayer(5),
            ]
        )

        self.play(Create(neural_network))

        self.play(
            neural_network.make_forward_pass_animation(
                layer_args={
                    embedding_layer: {
                        "triplet_args": {
                            "anchor_dist": {
                                "cov": np.array([[0.3, 0], [0, 0.3]]),
                                "mean": np.array([0.7, 1.4]),
                                "dist_theme": "ellipse",
                                "color": BLUE,
                            },
                            "positive_dist": {
                                "cov": np.array([[0.2, 0], [0, 0.2]]),
                                "mean": np.array([0.8, -0.4]),
                                "dist_theme": "ellipse",
                                "color": GREEN,
                            },
                            "negative_dist": {
                                "cov": np.array([[0.4, 0], [0, 0.25]]),
                                "mean": np.array([-1, -1.2]),
                                "dist_theme": "ellipse",
                                "color": RED,
                            },
                        }
                    }
                },
                run_time=5,
            )
        )


class QueryEmbeddingNNScene(Scene):
    def construct(self):
        embedding_layer = EmbeddingLayer()
        embedding_layer.paired_query_mode = True

        neural_network = NeuralNetwork(
            [
                FeedForwardLayer(5),
                FeedForwardLayer(3),
                embedding_layer,
                FeedForwardLayer(3),
                FeedForwardLayer(5),
            ]
        )

        self.play(Create(neural_network), run_time=2)

        self.play(
            neural_network.make_forward_pass_animation(
                run_time=5,
                layer_args={
                    embedding_layer: {
                        "positive_dist_args": {
                            "cov": np.array([[1, 0], [0, 1]]),
                            "mean": np.array([1, 1]),
                            "dist_theme": "ellipse",
                            "color": GREEN,
                        },
                        "negative_dist_args": {
                            "cov": np.array([[1, 0], [0, 1]]),
                            "mean": np.array([-1, -1]),
                            "dist_theme": "ellipse",
                            "color": RED,
                        },
                    }
                },
            )
        )



from manim import *
from manim_ml.neural_network.animations.dropout import (
    make_neural_network_dropout_animation,
)
from manim_ml.neural_network.layers.feed_forward import FeedForwardLayer
from manim_ml.neural_network.layers.image import ImageLayer
from PIL import Image
from manim_ml.neural_network.neural_network import NeuralNetwork
import numpy as np

config.pixel_height = 1200
config.pixel_width = 1900
config.frame_height = 5.0
config.frame_width = 5.0


def make_code_snippet():
    code_str = """ 
        nn = NeuralNetwork([
            FeedForwardLayer(3),
            FeedForwardLayer(5),
            FeedForwardLayer(3),
            FeedForwardLayer(5),
            FeedForwardLayer(4),
        ])
        self.play(
            make_neural_network_dropout_animation(
                nn, dropout_rate=0.25, do_forward_pass=True
            )
        )
    """

    code = Code(
        code=code_str,
        tab_width=4,
        background_stroke_width=1,
        background_stroke_color=WHITE,
        insert_line_no=False,
        style="monokai",
        language="py",
    )
    code.scale(0.28)

    return code


class DropoutNeuralNetworkScene(Scene):
    def construct(self):
        # Make nn
        nn = NeuralNetwork(
            [
                FeedForwardLayer(3, rectangle_color=BLUE),
                FeedForwardLayer(5, rectangle_color=BLUE),
                FeedForwardLayer(3, rectangle_color=BLUE),
                FeedForwardLayer(5, rectangle_color=BLUE),
                FeedForwardLayer(4, rectangle_color=BLUE),
            ],
            layer_spacing=0.4,
        )
        # Center the nn
        nn.move_to(ORIGIN)
        self.add(nn)
        # Make code snippet
        code_snippet = make_code_snippet()
        self.add(code_snippet)
        code_snippet.next_to(nn, DOWN * 0.7)
        Group(code_snippet, nn).move_to(ORIGIN)
        # Play animation
        self.play(
            make_neural_network_dropout_animation(
                nn, dropout_rate=0.25, do_forward_pass=True
            )
        )
        self.wait(1)


if __name__ == "__main__":
    """Render all scenes"""
    dropout_nn_scene = DropoutNeuralNetworkScene()
    dropout_nn_scene.render()



from manim import *


class ListGroup(Mobject):
    """Indexable Group with traditional list operations"""

    def __init__(self, *layers):
        super().__init__()
        self.items = [*layers]

    def __getitem__(self, indices):
        """Traditional list indexing"""
        return self.items[indices]

    def insert(self, index, item):
        """Inserts item at index"""
        self.items.insert(index, item)
        self.submobjects = self.items

    def remove_at_index(self, index):
        """Removes item at index"""
        if index > len(self.items):
            raise Exception(f"ListGroup index out of range: {index}")
        item = self.items[index]
        del self.items[index]
        self.submobjects = self.items

        return item

    def remove_at_indices(self, indices):
        """Removes items at indices"""
        items = []
        for index in indices:
            item = self.remove_at_index(index)
            items.append(item)

        return items

    def remove(self, item):
        """Removes first instance of item"""
        self.items.remove(item)
        self.submobjects = self.items

        return item

    def get(self, index):
        """Gets item at index"""
        return self.items[index]

    def add(self, item):
        """Adds to end"""
        self.items.append(item)
        self.submobjects = self.items

    def replace(self, index, item):
        """Replaces item at index"""
        self.items[index] = item
        self.submobjects = self.items

    def index_of(self, item):
        """Returns index of item if it exists"""
        for index, obj in enumerate(self.items):
            if item is obj:
                return index
        return -1

    def __len__(self):
        """Length of items"""
        return len(self.items)

    def set_z_index(self, z_index_value, family=True):
        """Sets z index of all values in ListGroup"""
        for item in self.items:
            item.set_z_index(z_index_value, family=True)

    def __iter__(self):
        self.current_index = -1
        return self

    def __next__(self):  # Python 2: def next(self)
        self.current_index += 1
        if self.current_index < len(self.items):
            return self.items[self.current_index]
        raise StopIteration

    def __repr__(self):
        return f"ListGroup({self.items})"



from manim_ml.neural_network.neural_network import NeuralNetwork
from manim_ml.neural_network.layers.feed_forward import FeedForwardLayer
from manim_ml.neural_network.layers.convolutional_2d_to_convolutional_2d import (
    Convolutional2DToConvolutional2D,
)
from manim_ml.neural_network.layers.convolutional_2d_to_feed_forward import (
    Convolutional2DToFeedForward,
)
from manim_ml.neural_network.layers.convolutional_2d_to_max_pooling_2d import (
    Convolutional2DToMaxPooling2D,
)
from manim_ml.neural_network.layers.convolutional_2d import Convolutional2DLayer
from manim_ml.neural_network.layers.embedding_to_feed_forward import (
    EmbeddingToFeedForward,
)
from manim_ml.neural_network.layers.embedding import EmbeddingLayer
from manim_ml.neural_network.layers.feed_forward_to_embedding import (
    FeedForwardToEmbedding,
)
from manim_ml.neural_network.layers.feed_forward_to_feed_forward import (
    FeedForwardToFeedForward,
)
from manim_ml.neural_network.layers.feed_forward_to_image import FeedForwardToImage
from manim_ml.neural_network.layers.feed_forward_to_vector import FeedForwardToVector
from manim_ml.neural_network.layers.feed_forward import FeedForwardLayer
from manim_ml.neural_network.layers.image_to_convolutional_2d import (
    ImageToConvolutional2DLayer,
)
from manim_ml.neural_network.layers.image_to_feed_forward import ImageToFeedForward
from manim_ml.neural_network.layers.image import ImageLayer
from manim_ml.neural_network.layers.max_pooling_2d_to_convolutional_2d import (
    MaxPooling2DToConvolutional2D,
)
from manim_ml.neural_network.layers.max_pooling_2d import MaxPooling2DLayer
from manim_ml.neural_network.layers.paired_query_to_feed_forward import (
    PairedQueryToFeedForward,
)
from manim_ml.neural_network.layers.paired_query import PairedQueryLayer
from manim_ml.neural_network.layers.triplet_to_feed_forward import TripletToFeedForward
from manim_ml.neural_network.layers.triplet import TripletLayer
from manim_ml.neural_network.layers.vector import VectorLayer
from manim_ml.neural_network.layers.math_operation_layer import MathOperationLayer


from manim import *


class SurroundingRectangleTest(Scene):
    def construct(self):
        rectangle = Rectangle(width=1, height=1, color=WHITE, fill_opacity=1.0)
        self.add(rectangle)

        surrounding_rectangle = SurroundingRectangle(rectangle, color=GREEN, buff=0.0)
        self.add(surrounding_rectangle)



"""Neural Network Manim Visualization

This module is responsible for generating a neural network visualization with
manim, specifically a fully connected neural network diagram.

Example:
    # Specify how many nodes are in each node layer
    layer_node_count = [5, 3, 5]
    # Create the object with default style settings
    NeuralNetwork(layer_node_count)
"""
import textwrap
from manim_ml.neural_network.layers.embedding import EmbeddingLayer
from manim_ml.utils.mobjects.connections import NetworkConnection
import numpy as np
from manim import *

from manim_ml.neural_network.layers.parent_layers import ConnectiveLayer, ThreeDLayer
from manim_ml.neural_network.layers.util import get_connective_layer
from manim_ml.utils.mobjects.list_group import ListGroup
from manim_ml.neural_network.animations.neural_network_transformations import (
    InsertLayer,
    RemoveLayer,
)
import manim_ml

class NeuralNetwork(Group):
    """Neural Network Visualization Container Class"""

    def __init__(
        self,
        input_layers,
        layer_spacing=0.2,
        animation_dot_color=manim_ml.config.color_scheme.active_color,
        edge_width=2.5,
        dot_radius=0.03,
        title=" ",
        layout="linear",
        layout_direction="left_to_right",
        debug_mode=False
    ):
        super(Group, self).__init__()
        self.input_layers_dict = self.make_input_layers_dict(input_layers)
        self.input_layers = ListGroup(*self.input_layers_dict.values())
        self.edge_width = edge_width
        self.layer_spacing = layer_spacing
        self.animation_dot_color = animation_dot_color
        self.dot_radius = dot_radius
        self.title_text = title
        self.created = False
        self.layout = layout
        self.layout_direction = layout_direction
        self.debug_mode = debug_mode
        # TODO take layer_node_count [0, (1, 2), 0]
        # and make it have explicit distinct subspaces
        # Construct all of the layers
        self._construct_input_layers()
        # Place the layers
        self._place_layers(layout=layout, layout_direction=layout_direction)
        # Make the connective layers
        self.connective_layers, self.all_layers = self._construct_connective_layers()
        # Place the connective layers
        self._place_connective_layers()
        # Make overhead title
        self.title = Text(
            self.title_text, 
            font_size=DEFAULT_FONT_SIZE / 2
        )
        self.title.next_to(self, UP, 1.0)
        self.add(self.title)
        # Place layers at correct z index
        self.connective_layers.set_z_index(2)
        self.input_layers.set_z_index(3)
        # Center the whole diagram by default
        self.all_layers.move_to(ORIGIN)
        self.add(self.all_layers)
        # Make container for connections
        self.connections = []
        # Print neural network
        print(repr(self))

    def make_input_layers_dict(self, input_layers):
        """Make dictionary of input layers"""
        if isinstance(input_layers, dict):
            # If input layers is dictionary then return it
            return input_layers
        elif isinstance(input_layers, list):
            # If input layers is a list then make a dictionary with default
            return_dict = {}
            for layer_index, input_layer in enumerate(input_layers):
                return_dict[f"layer{layer_index}"] = input_layer

            return return_dict
        else:
            raise Exception(f"Uncrecognized input layers type: {type(input_layers)}")

    def add_connection(
        self,
        start_mobject_or_name,
        end_mobject_or_name,
        connection_style="default",
        connection_position="bottom",
        arc_direction="down"
    ):
        """Add connection from start layer to end layer"""
        assert connection_style in ["default"]
        if connection_style == "default":
            # Make arrow connection from start layer to end layer
            # Add the connection
            if isinstance(start_mobject_or_name, Mobject):
                input_mobject = start_mobject_or_name
            else:
                input_mobject = self.input_layers_dict[start_mobject_or_name]
            if isinstance(end_mobject_or_name, Mobject):
                output_mobject = end_mobject_or_name
            else:
                output_mobject = self.input_layers_dict[end_mobject_or_name]
            
            connection = NetworkConnection(
                input_mobject,
                output_mobject,
                arc_direction=arc_direction,
                buffer=0.05
            )
            self.connections.append(connection)
            self.add(connection)

    def _construct_input_layers(self):
        """Constructs each of the input layers in context
        of their adjacent layers"""
        prev_layer = None
        next_layer = None
        # Go through all the input layers and run their construct method
        print("Constructing layers")
        for layer_index in range(len(self.input_layers)):
            current_layer = self.input_layers[layer_index]
            print(f"Current layer: {current_layer}")
            if layer_index < len(self.input_layers) - 1:
                next_layer = self.input_layers[layer_index + 1]
            if layer_index > 0:
                prev_layer = self.input_layers[layer_index - 1]
            # Run the construct layer method for each
            current_layer.construct_layer(
                prev_layer, 
                next_layer,
                debug_mode=self.debug_mode
            )

    def _place_layers(
        self, 
        layout="linear", 
        layout_direction="top_to_bottom"
    ):
        """Creates the neural network"""
        # TODO implement more sophisticated custom layouts
        # Default: Linear layout
        for layer_index in range(1, len(self.input_layers)):
            previous_layer = self.input_layers[layer_index - 1]
            current_layer = self.input_layers[layer_index]
            current_layer.move_to(previous_layer.get_center())

            if layout_direction == "left_to_right":
                x_shift = (
                    previous_layer.get_width() / 2
                    + current_layer.get_width() / 2
                    + self.layer_spacing
                )
                shift_vector = np.array([x_shift, 0, 0])
            elif layout_direction == "top_to_bottom":
                y_shift = -(
                    (previous_layer.get_width() / 2 + current_layer.get_width() / 2)
                    + self.layer_spacing
                )
                shift_vector = np.array([0, y_shift, 0])
            else:
                raise Exception(f"Unrecognized layout direction: {layout_direction}")
            current_layer.shift(shift_vector)

        # After all layers have been placed place their activation functions
        layer_max_height = max([layer.get_height() for layer in self.input_layers])
        for current_layer in self.input_layers:
            # Place activation function
            if hasattr(current_layer, "activation_function"):
                if not current_layer.activation_function is None:
                    # Get max height of layer
                    up_movement = np.array([
                        0,
                        layer_max_height / 2
                        + current_layer.activation_function.get_height() / 2
                        + 0.5 * self.layer_spacing,
                        0,
                    ])
                    current_layer.activation_function.move_to(
                        current_layer,
                    )
                    current_layer.activation_function.match_y(
                        self.input_layers[0]
                    )
                    current_layer.activation_function.shift(
                        up_movement
                    )
                    self.add(
                        current_layer.activation_function
                    )

    def _construct_connective_layers(self):
        """Draws connecting lines between layers"""
        connective_layers = ListGroup()
        all_layers = ListGroup()
        for layer_index in range(len(self.input_layers) - 1):
            current_layer = self.input_layers[layer_index]
            # Add the layer to the list of layers
            all_layers.add(current_layer)
            next_layer = self.input_layers[layer_index + 1]
            # Check if layer is actually a nested NeuralNetwork
            if isinstance(current_layer, NeuralNetwork):
                # Last layer of the current layer
                current_layer = current_layer.all_layers[-1]
            if isinstance(next_layer, NeuralNetwork):
                # First layer of the next layer
                next_layer = next_layer.all_layers[0]
            # Find connective layer with correct layer pair
            connective_layer = get_connective_layer(current_layer, next_layer)
            connective_layers.add(connective_layer)
            # Construct the connective layer
            connective_layer.construct_layer(current_layer, next_layer)
            # Add the layer to the list of layers
            all_layers.add(connective_layer)
        # Add final layer
        all_layers.add(self.input_layers[-1])
        # Handle layering
        return connective_layers, all_layers

    def _place_connective_layers(self):
        """Places the connective layers
        """
        # Place each of the connective layers halfway between the adjacent layers
        for connective_layer in self.connective_layers:
            layer_midpoint = (
                connective_layer.input_layer.get_center() +
                connective_layer.output_layer.get_center()
            ) / 2
            connective_layer.move_to(layer_midpoint)

    def insert_layer(self, layer, insert_index):
        """Inserts a layer at the given index"""
        neural_network = self
        insert_animation = InsertLayer(layer, insert_index, neural_network)
        return insert_animation

    def remove_layer(self, layer):
        """Removes layer object if it exists"""
        neural_network = self
        return RemoveLayer(layer, neural_network, layer_spacing=self.layer_spacing)

    def replace_layer(self, old_layer, new_layer):
        """Replaces given layer object"""
        raise NotImplementedError()
        remove_animation = self.remove_layer(insert_index)
        insert_animation = self.insert_layer(layer, insert_index)
        # Make the animation
        animation_group = AnimationGroup(
            FadeOut(self.all_layers[insert_index]), FadeIn(layer), lag_ratio=1.0
        )

        return animation_group

    def make_forward_pass_animation(
        self, 
        run_time=None, 
        passing_flash=True, 
        layer_args={}, 
        per_layer_animations=False,
        **kwargs
    ):
        """Generates an animation for feed forward propagation"""
        all_animations = []
        per_layer_animation_map = {}
        per_layer_runtime = (
            run_time / len(self.all_layers) if not run_time is None else None
        )
        for layer_index, layer in enumerate(self.all_layers):
            # Get the layer args
            if isinstance(layer, ConnectiveLayer):
                """
                NOTE: By default a connective layer will get the combined
                layer_args of the layers it is connecting and itself.
                """
                before_layer_args = {}
                current_layer_args = {}
                after_layer_args = {}
                if layer.input_layer in layer_args:
                    before_layer_args = layer_args[layer.input_layer]
                if layer in layer_args:
                    current_layer_args = layer_args[layer]
                if layer.output_layer in layer_args:
                    after_layer_args = layer_args[layer.output_layer]
                # Merge the two dicts
                current_layer_args = {
                    **before_layer_args,
                    **current_layer_args,
                    **after_layer_args,
                }
            else:
                current_layer_args = {}
                if layer in layer_args:
                    current_layer_args = layer_args[layer]
            # Perform the forward pass of the current layer
            layer_forward_pass = layer.make_forward_pass_animation(
                layer_args=current_layer_args, run_time=per_layer_runtime, **kwargs
            )
            # Animate a forward pass for incoming connections
            connection_input_pass = AnimationGroup()
            for connection in self.connections:
                if isinstance(layer, ConnectiveLayer):
                    output_layer = layer.output_layer
                    if connection.end_mobject == output_layer:
                        connection_input_pass = ShowPassingFlash(
                            connection,
                            run_time=layer_forward_pass.run_time,
                            time_width=0.2,
                        )
                        break

            layer_forward_pass = AnimationGroup(
                layer_forward_pass, 
                connection_input_pass, 
                lag_ratio=0.0
            )
            all_animations.append(layer_forward_pass)
            # Add the animation to per layer animation
            per_layer_animation_map[layer] = layer_forward_pass
        # Make the animation group
        animation_group = Succession(*all_animations, lag_ratio=1.0)
        if per_layer_animations:
            return per_layer_animation_map
        else:
            return animation_group

    @override_animation(Create)
    def _create_override(self, **kwargs):
        """Overrides Create animation"""
        # Stop the neural network from being created twice
        if self.created:
            return AnimationGroup()
        self.created = True

        animations = []
        # Create the overhead title
        animations.append(Create(self.title))
        # Create each layer one by one
        for layer in self.all_layers:
            layer_animation = Create(layer)
            # Make titles
            create_title = Create(layer.title)
            # Create layer animation group
            animation_group = AnimationGroup(layer_animation, create_title)
            animations.append(animation_group)

        animation_group = AnimationGroup(*animations, lag_ratio=1.0)

        return animation_group

    def set_z_index(self, z_index_value: float, family=False):
        """Overriden set_z_index"""
        # Setting family=False stops sub-neural networks from inheriting parent z_index
        for layer in self.all_layers:
            if not isinstance(NeuralNetwork):
                layer.set_z_index(z_index_value)

    def scale(self, scale_factor, **kwargs):
        """Overriden scale"""
        prior_center = self.get_center()

        for layer in self.all_layers:
            layer.scale(scale_factor, **kwargs)
        # Place layers with scaled spacing
        self.layer_spacing *= scale_factor
        # self.connective_layers, self.all_layers = self._construct_connective_layers()
        self._place_layers(layout=self.layout, layout_direction=self.layout_direction)
        self._place_connective_layers()
        # Scale the title
        self.remove(self.title)
        self.title.scale(scale_factor, **kwargs)
        self.title.next_to(self, UP * scale_factor, buff=1.0 * scale_factor)
        self.add(self.title)

        self.move_to(prior_center)

    def filter_layers(self, function):
        """Filters layers of the network given function"""
        layers_to_return = []
        for layer in self.all_layers:
            func_out = function(layer)
            assert isinstance(
                func_out, bool
            ), "Filter layers function returned a non-boolean type."
            if func_out:
                layers_to_return.append(layer)

        return layers_to_return

    def __repr__(self, metadata=["z_index", "title_text"]):
        """Print string representation of layers"""
        inner_string = ""
        for layer in self.all_layers:
            inner_string += f"{repr(layer)}("
            for key in metadata:
                value = getattr(layer, key)
                if not value is "":
                    inner_string += f"{key}={value}, "
            inner_string += "),\n"
        inner_string = textwrap.indent(inner_string, "    ")

        string_repr = "NeuralNetwork([\n" + inner_string + "])"
        return string_repr



"""Variational Autoencoder Manim Visualizations

In this module I define Manim visualizations for Variational Autoencoders
and Traditional Autoencoders.

"""
from manim import *
import numpy as np
from PIL import Image
from manim_ml.neural_network.layers import FeedForwardLayer, EmbeddingLayer, ImageLayer
from manim_ml.neural_network.neural_network import NeuralNetwork


class VariationalAutoencoder(VGroup):
    """Variational Autoencoder Manim Visualization"""

    def __init__(
        self,
        encoder_nodes_per_layer=[5, 3],
        decoder_nodes_per_layer=[3, 5],
        point_color=BLUE,
        dot_radius=0.05,
        ellipse_stroke_width=1.0,
        layer_spacing=0.5,
    ):
        super(VGroup, self).__init__()
        self.encoder_nodes_per_layer = encoder_nodes_per_layer
        self.decoder_nodes_per_layer = decoder_nodes_per_layer
        self.point_color = point_color
        self.dot_radius = dot_radius
        self.layer_spacing = layer_spacing
        self.ellipse_stroke_width = ellipse_stroke_width
        # Make the VMobjects
        self.neural_network, self.embedding_layer = self._construct_neural_network()

    def _construct_neural_network(self):
        """Makes the VAE encoder, embedding layer, and decoder"""
        embedding_layer = EmbeddingLayer()

        neural_network = NeuralNetwork(
            [
                FeedForwardLayer(5),
                FeedForwardLayer(3),
                embedding_layer,
                FeedForwardLayer(3),
                FeedForwardLayer(5),
            ]
        )

        return neural_network, embedding_layer

    @override_animation(Create)
    def _create_vae(self):
        return Create(self.neural_network)

    def make_triplet_forward_pass(self, triplet):
        pass

    def make_image_forward_pass(self, input_image, output_image, run_time=1.5):
        """Override forward pass animation specific to a VAE"""
        # Make a wrapper NN with images
        wrapper_neural_network = NeuralNetwork(
            [ImageLayer(input_image), self.neural_network, ImageLayer(output_image)]
        )
        # Make animation
        animation_group = AnimationGroup(
            Create(wrapper_neural_network),
            wrapper_neural_network.make_forward_pass_animation(),
            lag_ratio=1.0,
        )

        return animation_group



from manim import *
from abc import ABC, abstractmethod
import random

import manim_ml.neural_network.activation_functions.relu as relu
import manim_ml

class ActivationFunction(ABC, VGroup):
    """Abstract parent class for defining activation functions"""

    def __init__(
        self,
        function_name=None,
        x_range=[-1, 1],
        y_range=[-1, 1],
        x_length=0.5,
        y_length=0.3,
        show_function_name=True,
        active_color=manim_ml.config.color_scheme.active_color,
        plot_color=manim_ml.config.color_scheme.primary_color,
        rectangle_color=manim_ml.config.color_scheme.secondary_color,
    ):
        super(VGroup, self).__init__()
        self.function_name = function_name
        self.x_range = x_range
        self.y_range = y_range
        self.x_length = x_length
        self.y_length = y_length
        self.show_function_name = show_function_name
        self.active_color = active_color
        self.plot_color = plot_color
        self.rectangle_color = rectangle_color

        self.construct_activation_function()

    def construct_activation_function(self):
        """Makes the activation function"""
        # Make an axis
        self.axes = Axes(
            x_range=self.x_range,
            y_range=self.y_range,
            x_length=self.x_length,
            y_length=self.y_length,
            tips=False,
            axis_config={
                "include_numbers": False,
                "stroke_width": 0.5,
                "include_ticks": False,
                "color": self.rectangle_color
            },
        )
        self.add(self.axes)
        # Surround the axis with a rounded rectangle.
        self.surrounding_rectangle = SurroundingRectangle(
            self.axes,
            corner_radius=0.05,
            buff=0.05,
            stroke_width=2.0,
            stroke_color=self.rectangle_color,
        )
        self.add(self.surrounding_rectangle)
        # Plot function on axis by applying it and showing in given range
        self.graph = self.axes.plot(
            lambda x: self.apply_function(x),
            x_range=self.x_range,
            stroke_color=self.plot_color,
            stroke_width=2.0,
        )
        self.add(self.graph)
        # Add the function name
        if self.show_function_name:
            function_name_text = Text(
                self.function_name, font_size=12, font="sans-serif"
            )
            function_name_text.next_to(self.axes, UP * 0.5)
            self.add(function_name_text)

    @abstractmethod
    def apply_function(self, x_val):
        """Evaluates function at given x_val"""
        if x_val == None:
            x_val = random.uniform(self.x_range[0], self.x_range[1])

    def make_evaluate_animation(self, x_val=None):
        """Evaluates the function at a random point in the x_range"""
        # Highlight the graph
        # TODO: Evaluate the function at the x_val and show a highlighted dot
        animation_group = Succession(
            AnimationGroup(
                ApplyMethod(self.graph.set_color, self.active_color),
                ApplyMethod(
                    self.surrounding_rectangle.set_stroke_color, self.active_color
                ),
                lag_ratio=0.0,
            ),
            Wait(1),
            AnimationGroup(
                ApplyMethod(self.graph.set_color, self.plot_color),
                ApplyMethod(
                    self.surrounding_rectangle.set_stroke_color, self.rectangle_color
                ),
                lag_ratio=0.0,
            ),
            lag_ratio=1.0,
        )

        return animation_group



from manim import *
from manim_ml.utils.mobjects.gridded_rectangle import GriddedRectangle


class TestGriddedRectangleScene(ThreeDScene):
    def construct(self):
        rect = GriddedRectangle(color=ORANGE, width=3, height=3)
        self.add(rect)



from manim import *
from manim_ml.neural_network.layers.feed_forward import FeedForwardLayer
from manim_ml.neural_network.neural_network import NeuralNetwork

import manim_ml

print(manim_ml.config.color_scheme)

class NeuralNetworkScene(Scene):
    def construct(self):
        layers = [FeedForwardLayer(3), FeedForwardLayer(5), FeedForwardLayer(3)]
        nn = NeuralNetwork(layers)
        self.add(nn)
        # nn.move_to(ORIGIN)
        self.wait(2)
        # nn.scale(2)
        forward_propagation_animation = nn.make_forward_pass_animation(
            run_time=5, passing_flash=True
        )
        self.play(forward_propagation_animation)


from cv2 import exp
from manim import *
from manim_ml.neural_network.layers.embedding import EmbeddingLayer
from manim_ml.neural_network.layers.embedding_to_feed_forward import (
    EmbeddingToFeedForward,
)
from manim_ml.neural_network.layers.feed_forward import FeedForwardLayer
from manim_ml.neural_network.layers.feed_forward_to_embedding import (
    FeedForwardToEmbedding,
)
from manim_ml.neural_network.layers.feed_forward_to_feed_forward import (
    FeedForwardToFeedForward,
)
from manim_ml.neural_network.layers.image import ImageLayer
from manim_ml.neural_network.neural_network import (
    NeuralNetwork,
    FeedForwardNeuralNetwork,
)
from PIL import Image
import numpy as np

config.pixel_height = 720
config.pixel_width = 1280
config.frame_height = 6.0
config.frame_width = 6.0

"""
    Unit Tests
"""


def assert_classes_match(all_layers, expected_classes):
    assert len(list(all_layers)) == 5

    for index, layer in enumerate(all_layers):
        expected_class = expected_classes[index]
        assert isinstance(
            layer, expected_class
        ), f"Wrong layer class {layer.__class__} expected {expected_class}"


def test_embedding_layer():
    embedding_layer = EmbeddingLayer()

    neural_network = NeuralNetwork(
        [FeedForwardLayer(5), FeedForwardLayer(3), embedding_layer]
    )

    expected_classes = [
        FeedForwardLayer,
        FeedForwardToFeedForward,
        FeedForwardLayer,
        FeedForwardToEmbedding,
        EmbeddingLayer,
    ]

    assert_classes_match(neural_network.all_layers, expected_classes)


def test_remove_layer():
    embedding_layer = EmbeddingLayer()

    neural_network = NeuralNetwork(
        [FeedForwardLayer(5), FeedForwardLayer(3), embedding_layer]
    )

    expected_classes = [
        FeedForwardLayer,
        FeedForwardToFeedForward,
        FeedForwardLayer,
        FeedForwardToEmbedding,
        EmbeddingLayer,
    ]

    assert_classes_match(neural_network.all_layers, expected_classes)

    print("before removal")
    print(list(neural_network.all_layers))
    neural_network.remove_layer(embedding_layer)
    print("after removal")
    print(list(neural_network.all_layers))

    expected_classes = [
        FeedForwardLayer,
        FeedForwardToFeedForward,
        FeedForwardLayer,
    ]

    print(list(neural_network.all_layers))

    assert_classes_match(neural_network.all_layers, expected_classes)


class FeedForwardNeuralNetworkScene(Scene):
    def construct(self):
        nn = FeedForwardNeuralNetwork([3, 5, 3])
        self.play(Create(nn))
        self.play(Wait(3))


class NeuralNetworkScene(Scene):
    """Test Scene for the Neural Network"""

    def construct(self):
        # Make the Layer object
        layers = [FeedForwardLayer(3), FeedForwardLayer(5), FeedForwardLayer(3)]
        nn = NeuralNetwork(layers)
        nn.move_to(ORIGIN)
        # Make Animation
        self.add(nn)
        # self.play(Create(nn))
        forward_propagation_animation = nn.make_forward_pass_animation(
            run_time=5, passing_flash=True
        )

        self.play(forward_propagation_animation)


class GrayscaleImageNeuralNetworkScene(Scene):
    def construct(self):
        image = Image.open("images/image.jpeg")
        numpy_image = np.asarray(image)
        # Make nn
        layers = [
            FeedForwardLayer(3),
            FeedForwardLayer(5),
            FeedForwardLayer(3),
            FeedForwardLayer(6),
            ImageLayer(numpy_image, height=1.4),
        ]
        nn = NeuralNetwork(layers)
        nn.scale(1.3)
        # Center the nn
        nn.move_to(ORIGIN)
        self.add(nn)
        # Play animation
        self.play(nn.make_forward_pass_animation(run_time=5))
        self.play(nn.make_forward_pass_animation(run_time=5))


class ImageNeuralNetworkScene(Scene):
    def construct(self):
        image = Image.open("../assets/gan/real_image.jpg")
        numpy_image = np.asarray(image)
        # Make nn
        layers = [
            FeedForwardLayer(3),
            FeedForwardLayer(5),
            FeedForwardLayer(3),
            FeedForwardLayer(6),
            ImageLayer(numpy_image, height=1.4),
        ]
        nn = NeuralNetwork(layers)
        nn.scale(1.3)
        # Center the nn
        nn.move_to(ORIGIN)
        self.add(nn)
        # Play animation
        self.play(nn.make_forward_pass_animation(run_time=5))
        self.play(nn.make_forward_pass_animation(run_time=5))


class RecursiveNNScene(Scene):
    def construct(self):
        nn = NeuralNetwork(
            [
                NeuralNetwork([FeedForwardLayer(3), FeedForwardLayer(2)]),
                NeuralNetwork([FeedForwardLayer(2), FeedForwardLayer(3)]),
            ]
        )

        self.play(Create(nn))


class LayerInsertionScene(Scene):
    def construct(self):
        pass


class LayerRemovalScene(Scene):
    def construct(self):
        image = Image.open("images/image.jpeg")
        numpy_image = np.asarray(image)

        layer = FeedForwardLayer(5)
        layers = [
            ImageLayer(numpy_image, height=1.4),
            FeedForwardLayer(3),
            layer,
            FeedForwardLayer(3),
            FeedForwardLayer(6),
        ]

        nn = NeuralNetwork(layers)

        self.play(Create(nn))
        remove_animation = nn.remove_layer(layer)
        print("before remove")
        self.play(remove_animation)
        print(nn)
        print("after remove")


class LayerInsertionScene(Scene):
    def construct(self):
        image = Image.open("images/image.jpeg")
        numpy_image = np.asarray(image)

        layers = [
            ImageLayer(numpy_image, height=1.4),
            FeedForwardLayer(3),
            FeedForwardLayer(3),
            FeedForwardLayer(6),
        ]

        nn = NeuralNetwork(layers)

        self.play(Create(nn))

        layer = FeedForwardLayer(5)
        insert_animation = nn.insert_layer(layer, 4)
        self.play(insert_animation)
        print(nn)
        print("after remove")


if __name__ == "__main__":
    """Render all scenes"""
    # Feed Forward Neural Network
    ffnn_scene = FeedForwardNeuralNetworkScene()
    ffnn_scene.render()
    # Neural Network
    nn_scene = NeuralNetworkScene()
    nn_scene.render()



from manim import *
from manim_ml.neural_network.layers.parent_layers import NeuralNetworkLayer
from manim_ml.utils.mobjects.image import GrayscaleImageMobject, LabeledColorImage
import numpy as np


class PairedQueryLayer(NeuralNetworkLayer):
    """Paired Query Layer"""

    def __init__(
        self, positive, negative, stroke_width=5, font_size=18, spacing=0.5, **kwargs
    ):
        super().__init__(**kwargs)
        self.positive = positive
        self.negative = negative
        self.font_size = font_size
        self.spacing = spacing

        self.stroke_width = stroke_width
        # Make the assets
        self.assets = self.make_assets()
        self.add(self.assets)
        self.add(self.title)

    def construct_layer(
        self,
        input_layer: "NeuralNetworkLayer",
        output_layer: "NeuralNetworkLayer",
        **kwargs
    ):
        return super().construct_layer(input_layer, output_layer, **kwargs)

    @classmethod
    def from_paths(cls, positive_path, negative_path, grayscale=True, **kwargs):
        """Creates a query using the paths"""
        # Load images from path
        if grayscale:
            positive = GrayscaleImageMobject.from_path(positive_path)
            negative = GrayscaleImageMobject.from_path(negative_path)
        else:
            positive = ImageMobject(positive_path)
            negative = ImageMobject(negative_path)
        # Make the layer
        query_layer = cls(positive, negative, **kwargs)

        return query_layer

    def make_assets(self):
        """
        Constructs the assets needed for a query layer
        """
        # Handle positive
        positive_group = LabeledColorImage(
            self.positive,
            color=BLUE,
            label="Positive",
            font_size=self.font_size,
            stroke_width=self.stroke_width,
        )
        # Handle negative
        negative_group = LabeledColorImage(
            self.negative,
            color=RED,
            label="Negative",
            font_size=self.font_size,
            stroke_width=self.stroke_width,
        )
        # Distribute the groups uniformly vertically
        assets = Group(positive_group, negative_group)
        assets.arrange(DOWN, buff=self.spacing)

        return assets

    @override_animation(Create)
    def _create_override(self):
        # TODO make Create animation that is custom
        return FadeIn(self.assets)

    def make_forward_pass_animation(self, layer_args={}, **kwargs):
        """Forward pass for query"""
        return AnimationGroup()



from transformers import AutoTokenizer, BertModel
from manim import *
import matplotlib


class SelfAttentionGraph(VGroup):
    """VGroup object for visualizing self attention scores."""

    def __init__(
        self,
        sentence,
        tokenizer,
        transformer,
        layer_index=0,
        head_index=0,
        device="cpu",
    ):
        """Takes a sentence and passes it through a given transformer model. 
        Saves each of the self attention scores for the given layer_id (with 
        a default being the first layer). Returns tensor of self attention 
        scores. The tensor is of shape (num_heads, seq_len, seq_len).

        Parameters
        ----------
        sentence : str
            Input text sentence
        tokenizer : transformers.Tokenizer
            Tokenizer to apply to the sentence
        transformer : transformers.PreTrainedModel
            Transformer model to pass the sentence through.
        layer_index : int, optional
            index of layer to pull attentions from, by default None
        device : str, optional
            device identifier string, by default "cpu"
        """
        super().__init__()
        self.sentence = sentence
        self.tokenizer = tokenizer
        self.transformer = transformer
        self.layer_index = layer_index
        self.head_index = head_index
        self.device = device
        # Compute the attention scores
        inputs = self.tokenizer(
            self.sentence, 
            return_tensors="pt"
        )
        self.tokens = self.tokenizer.convert_ids_to_tokens(
            inputs["input_ids"][0]
        )
        outputs = self.transformer(
            **inputs,
            output_attentions=True,
        )
        self.attention_scores = outputs.attentions[self.layer_index].squeeze()
        # Make the attention graph
        attention_graph = self.make_attention_graph()
        self.add(attention_graph)

    def make_attention_graph(self, token_spacing=0.6, side_spacing=4.0, cmap="inferno"):
        """Creates the graph of the attention scores"""
        graph = VGroup()
        left_tokens = self.tokens.copy()
        right_tokens = self.tokens.copy()
        left_text = []
        right_text = []
        # Plot the edges
        for token in left_tokens:
            left_text.append(
                Text(token)
            )
            if len(left_text) > 1:
                left_text[-1].next_to(
                    left_text[-2], 
                    DOWN, 
                    buff=token_spacing
                )
        left_tokens = VGroup(*left_text)

        for token in right_tokens:
            right_text.append(
                Text(token)
            )
            if len(right_text) > 1:
                right_text[-1].next_to(
                    right_text[-2], 
                    DOWN, 
                    buff=token_spacing
                )
        right_tokens = VGroup(*right_text)
        right_tokens.shift(
            side_spacing * RIGHT
        )
        # Plot the edges
        edges = VGroup()
        for token_index, token in enumerate(self.tokens):
            for other_token_index, other_token in enumerate(self.tokens):
                edge = Line(
                    left_tokens[token_index].get_center(),
                    right_tokens[other_token_index].get_center(),
                    stroke_width=4.0,
                )
                attention_score = self.attention_scores[self.head_index, token_index, other_token_index]
                cmap = matplotlib.cm.get_cmap(cmap)
                rgba = np.array(cmap(
                    attention_score.item()
                ))
                rgb = (rgba * 255).astype(int)
                hex = '#%02x%02x%02x' % (rgb[0], rgb[1], rgb[2])

                edge.set_color(
                    hex
                )
                edges.add(edge)
        graph.add(edges)
        graph.add(left_tokens, right_tokens)

        return graph

class SelfAttentionGraphTest(Scene):

    def construct(self):
        tokenizer = AutoTokenizer.from_pretrained("bert-base-uncased")
        model = BertModel.from_pretrained("bert-base-uncased")

        self_attention_graph = SelfAttentionGraph(
            "This is a test sentence",
            tokenizer,
            model,
        )
        self_attention_graph.move_to(ORIGIN)
        self.add(self_attention_graph)
    


from manim import *
from PIL import Image
from manim_ml.neural_network.layers.feed_forward import FeedForwardLayer
from manim_ml.neural_network.neural_network import NeuralNetwork
import numpy as np

config.pixel_height = 720
config.pixel_width = 1280
config.frame_height = 6.0
config.frame_width = 6.0

class NeuralNetworkScene(Scene):
    """Test Scene for the Neural Network"""

    def construct(self):
        # Make the Layer object
        nn = NeuralNetwork([
            FeedForwardLayer(3),
            FeedForwardLayer(5),
            FeedForwardLayer(3),
        ] , title="Test Title")
        nn.move_to(ORIGIN)
        # Make Animation
        self.add(nn)
        nn.shift(RIGHT)
        nn.move_to(ORIGIN)
        self.wait(2)
        nn.scale(0.5)
        # self.play(Create(nn))
        forward_propagation_animation = nn.make_forward_pass_animation(
            run_time=5, passing_flash=True
        )

        self.play(forward_propagation_animation)



from manim import *
from manim_ml.neural_network.layers import PairedQueryLayer
from manim_ml.neural_network.layers.feed_forward import FeedForwardLayer
from manim_ml.neural_network.neural_network import NeuralNetwork

config.pixel_height = 720
config.pixel_width = 1280
config.frame_height = 6.0
config.frame_width = 6.0


class PairedQueryScene(Scene):
    def construct(self):
        positive_path = "../assets/triplet/positive.jpg"
        negative_path = "../assets/triplet/negative.jpg"

        paired_layer = PairedQueryLayer.from_paths(
            positive_path, negative_path, grayscale=False
        )

        paired_layer.scale(0.08)

        neural_network = NeuralNetwork(
            [paired_layer, FeedForwardLayer(5), FeedForwardLayer(3)]
        )

        neural_network.scale(1)

        self.play(Create(neural_network), run_time=3)

        self.play(neural_network.make_forward_pass_animation(), run_time=10)



from PIL import Image

from manim import *
from manim_ml.utils.mobjects.image import GrayscaleImageMobject
from manim_ml.neural_network.layers.parent_layers import ThreeDLayer
import manim_ml


class TestImageHomotopy(Scene):
    def compute_shape_at_time(self, time):
        """Computes the shape of a transformed image at a given time"""

    def compute_center_offset_at_time(self, x, y, z, time):
        """Computes the center offset of a point at a given time"""
        pass

    def construct(self):
        image = Image.open("../assets/mnist/digit.jpeg")
        numpy_image = np.asarray(image)
        # Make nn
        image_mobject = GrayscaleImageMobject(numpy_image)
        self.add(image_mobject)
        self.wait(1)

        rot = image_mobject.animate.rotate(
            axis=[0, 1, 0], 
            angle=manim_ml.config.three_d_config.three_d_y_rotation
        )
        move = image_mobject.animate.move_to()
        self.play(rot)
        """
        # Make square
        square = Square()
        self.add(square)
        # Make polygon
        polygon = Polygon(
            [1, 0, 0],
            [0, 1, 0],
            [-1, 0, 0],
            [0, -1, 0],
        )
        polygon.shift(RIGHT)
        self.play(
            Transform(square, polygon)
        )
        # Make the homotopy

        def shift_right_homotopy(x, y, z, t):
            return x + 1, y, z
        # Make the animation
        animation = Homotopy(
            mobject=image_mobject,
            homotopy=shift_right_homotopy
        )

        self.play(animation, run_time=1)
        """



from manim import *

from PIL import Image
import numpy as np

class NeuralNetworkScene(ThreeDScene):
    """Test Scene for the Neural Network"""

    def construct(self):
        # Make the Layer object
        image = Image.open("../assets/mnist/digit.jpeg")
        numpy_image = np.asarray(image)

        square = Square().shift([0.0, 2.5, 0.0])
        self.add(square)

        image_mobject = ImageMobject(
            numpy_image, 
            image_mode="RGB"
        ).scale(10)

        self.add(image_mobject)
        self.play(
            self.camera._frame_center.animate.shift(
                np.array([3.0, 0.0, 0])
            )
        )


from manim import *
from manim_ml.utils.mobjects.gridded_rectangle import GriddedRectangle

from manim_ml.neural_network.layers.parent_layers import (
    ThreeDLayer,
    VGroupNeuralNetworkLayer,
)
import manim_ml

class MaxPooling2DLayer(VGroupNeuralNetworkLayer, ThreeDLayer):
    """Max pooling layer for Convolutional2DLayer

    Note: This is for a Convolutional2DLayer even though
    it is called MaxPooling2DLayer because the 2D corresponds
    to the 2 spatial dimensions of the convolution.
    """

    def __init__(
        self,
        kernel_size=2,
        stride=1,
        cell_highlight_color=ORANGE,
        cell_width=0.2,
        filter_spacing=0.1,
        color=BLUE,
        show_grid_lines=False,
        stroke_width=2.0,
        **kwargs
    ):
        """Layer object for animating 2D Convolution Max Pooling

        Parameters
        ----------
        kernel_size : int or tuple, optional
            Width/Height of max pooling kernel, by default 2
        stride : int, optional
            Stride of the max pooling operation, by default 1
        """
        super().__init__(**kwargs)
        self.kernel_size = kernel_size
        self.stride = stride
        self.cell_highlight_color = cell_highlight_color
        self.cell_width = cell_width
        self.filter_spacing = filter_spacing
        self.color = color
        self.show_grid_lines = show_grid_lines
        self.stroke_width = stroke_width
        self.padding = (0, 0)

    def construct_layer(
        self,
        input_layer: "NeuralNetworkLayer",
        output_layer: "NeuralNetworkLayer",
        **kwargs
    ):
        # Make the output feature maps
        self.feature_maps = self._make_output_feature_maps(
            input_layer.num_feature_maps, input_layer.feature_map_size
        )
        self.add(self.feature_maps)
        self.rotate(
            manim_ml.config.three_d_config.rotation_angle,
            about_point=self.get_center(),
            axis=manim_ml.config.three_d_config.rotation_axis
        )
        self.feature_map_size = (
            input_layer.feature_map_size[0] / self.kernel_size,
            input_layer.feature_map_size[1] / self.kernel_size,
        )
        super().construct_layer(input_layer, output_layer, **kwargs)


    def _make_output_feature_maps(self, num_input_feature_maps, input_feature_map_size):
        """Makes a set of output feature maps"""
        # Compute the size of the feature maps
        output_feature_map_size = (
            input_feature_map_size[0] / self.kernel_size,
            input_feature_map_size[1] / self.kernel_size,
        )
        # Draw rectangles that are filled in with opacity
        feature_maps = []
        for filter_index in range(num_input_feature_maps):
            rectangle = GriddedRectangle(
                color=self.color,
                height=output_feature_map_size[1] * self.cell_width,
                width=output_feature_map_size[0] * self.cell_width,
                fill_color=self.color,
                fill_opacity=0.2,
                stroke_color=self.color,
                stroke_width=self.stroke_width,
                grid_xstep=self.cell_width,
                grid_ystep=self.cell_width,
                grid_stroke_width=self.stroke_width / 2,
                grid_stroke_color=self.color,
                show_grid_lines=self.show_grid_lines,
            )
            # Move the feature map
            rectangle.move_to([0, 0, filter_index * self.filter_spacing])
            # rectangle.set_z_index(4)
            feature_maps.append(rectangle)

        return VGroup(*feature_maps)

    def make_forward_pass_animation(self, layer_args={}, **kwargs):
        """Makes forward pass of Max Pooling Layer.

        Parameters
        ----------
        layer_args : dict, optional
            _description_, by default {}
        """
        return AnimationGroup()

    @override_animation(Create)
    def _create_override(self, **kwargs):
        """Create animation for the MaxPooling operation"""
        pass



from manim import *
from manim_ml.utils.testing.frames_comparison import frames_comparison

from manim_ml.neural_network import NeuralNetwork, FeedForwardLayer

__module_test__ = "feed_forward"


@frames_comparison
def test_FeedForwardScene(scene):
    """Tests the appearance of a feed forward network"""
    nn = NeuralNetwork([FeedForwardLayer(3), FeedForwardLayer(5), FeedForwardLayer(3)])

    scene.add(nn)


class FeedForwardScene(Scene):
    def construct(self):
        nn = NeuralNetwork(
            [FeedForwardLayer(3), FeedForwardLayer(5), FeedForwardLayer(3)]
        )

        self.add(nn)

        self.play(nn.make_forward_pass_animation())



from manim import *
import numpy as np

from manim_ml.neural_network.activation_functions.activation_function import (
    ActivationFunction,
)


class SigmoidFunction(ActivationFunction):
    """Sigmoid Activation Function"""

    def __init__(self, function_name="Sigmoid", x_range=[-5, 5], y_range=[0, 1]):
        super().__init__(function_name, x_range, y_range)

    def apply_function(self, x_val):
        return 1 / (1 + np.exp(-1 * x_val))



from manim import *


class TestSuccession(Scene):
    def construct(self):
        white_dot = Dot(color=WHITE)
        white_dot.shift(UP)

        red_dot = Dot(color=RED)

        self.play(
            Succession(
                Create(white_dot),
                white_dot.animate.shift(RIGHT),
                Create(red_dot),
                Wait(1),
                Uncreate(red_dot),
            )
        )



from manim import *
import numpy as np
from PIL import Image

from manim_ml.utils.mobjects.image import GrayscaleImageMobject
from manim_ml.neural_network.layers.parent_layers import NeuralNetworkLayer


class ImageLayer(NeuralNetworkLayer):
    """Single Image Layer for Neural Network"""

    def __init__(
        self, 
        numpy_image, 
        height=1.5, 
        show_image_on_create=True, 
        **kwargs
    ):
        super().__init__(**kwargs)
        self.image_height = height
        self.numpy_image = numpy_image
        self.show_image_on_create = show_image_on_create

    def construct_layer(self, input_layer, output_layer, **kwargs):
        """Construct layer method

        Parameters
        ----------
        input_layer :
            Input layer
        output_layer :
            Output layer
        """
        if len(np.shape(self.numpy_image)) == 2:
            # Assumed Grayscale
            self.num_channels = 1
            self.image_mobject = GrayscaleImageMobject(
                self.numpy_image, 
                height=self.image_height
            )
        elif len(np.shape(self.numpy_image)) == 3:
            # Assumed RGB
            self.num_channels = 3
            self.image_mobject = ImageMobject(self.numpy_image).scale_to_fit_height(
                self.image_height
            )
        self.add(self.image_mobject)
        super().construct_layer(input_layer, output_layer, **kwargs)

    @classmethod
    def from_path(cls, image_path, grayscale=True, **kwargs):
        """Creates a query using the paths"""
        # Load images from path
        image = Image.open(image_path)
        numpy_image = np.asarray(image)
        # Make the layer
        image_layer = cls(numpy_image, **kwargs)

        return image_layer

    @override_animation(Create)
    def _create_override(self, **kwargs):
        debug_mode = False
        if debug_mode:
            return FadeIn(SurroundingRectangle(self.image_mobject))
        if self.show_image_on_create:
            return FadeIn(self.image_mobject)
        else:
            return AnimationGroup()

    def make_forward_pass_animation(self, layer_args={}, **kwargs):
        return AnimationGroup()

    def get_right(self):
        """Override get right"""
        return self.image_mobject.get_right()

    def scale(self, scale_factor, **kwargs):
        """Scales the image mobject"""
        self.image_mobject.scale(scale_factor)

    @property
    def width(self):
        return self.image_mobject.width

    @property
    def height(self):
        return self.image_mobject.height



from manim import *

from manim_ml.neural_network import NeuralNetwork, FeedForwardLayer
import manim_ml

manim_ml.config.color_scheme = "light_mode"

config.pixel_height = 1200
config.pixel_width = 1900
config.frame_height = 6.0
config.frame_width = 6.0

class FeedForwardScene(Scene):
    def construct(self):
        nn = NeuralNetwork([
            FeedForwardLayer(3), 
            FeedForwardLayer(5), 
            FeedForwardLayer(3)
        ])

        self.add(nn)
        self.play(nn.make_forward_pass_animation())



from manim import *
import numpy as np
import math


class GaussianDistribution(VGroup):
    """Object for drawing a Gaussian distribution"""

    def __init__(
        self, axes, mean=None, cov=None, dist_theme="gaussian", color=ORANGE, **kwargs
    ):
        super(VGroup, self).__init__(**kwargs)
        self.axes = axes
        self.mean = mean
        self.cov = cov
        self.dist_theme = dist_theme
        self.color = color
        if mean is None:
            self.mean = np.array([0.0, 0.0])
        if cov is None:
            self.cov = np.array([[1, 0], [0, 1]])
        # Make the Gaussian
        if self.dist_theme is "gaussian":
            self.ellipses = self.construct_gaussian_distribution(
                self.mean, self.cov, color=self.color
            )
            self.add(self.ellipses)
        elif self.dist_theme is "ellipse":
            self.ellipses = self.construct_simple_gaussian_ellipse(
                self.mean, self.cov, color=self.color
            )
            self.add(self.ellipses)
        else:
            raise Exception(f"Uncrecognized distribution theme: {self.dist_theme}")

    """  
    @override_animation(Create)
    def _create_gaussian_distribution(self):
        return Create(self)
    """

    def compute_covariance_rotation_and_scale(self, covariance):
        def eigsorted(cov):
            """
            Eigenvalues and eigenvectors of the covariance matrix.
            """
            vals, vecs = np.linalg.eigh(cov)
            order = vals.argsort()[::-1]
            return vals[order], vecs[:, order]

        def cov_ellipse(cov, nstd):
            """
            Source: http://stackoverflow.com/a/12321306/1391441
            """

            vals, vecs = eigsorted(cov)
            theta = np.degrees(np.arctan2(*vecs[:, 0][::-1]))

            # Width and height are "full" widths, not radius
            width, height = 2 * nstd * np.sqrt(vals)

            return width, height, theta

        width, height, angle = cov_ellipse(covariance, 1)
        scale_factor = (
            np.abs(self.axes.x_range[0] - self.axes.x_range[1]) / self.axes.x_length
        )
        width /= scale_factor
        height /= scale_factor
        return angle, width, height

    def construct_gaussian_distribution(
        self, mean, covariance, color=ORANGE, num_ellipses=4
    ):
        """Returns a 2d Gaussian distribution object with given mean and covariance"""
        # map mean and covariance to frame coordinates
        mean = self.axes.coords_to_point(*mean)
        # Figure out the scale and angle of rotation
        rotation, width, height = self.compute_covariance_rotation_and_scale(covariance)
        # Make covariance ellipses
        opacity = 0.0
        ellipses = VGroup()
        for ellipse_number in range(num_ellipses):
            opacity += 1.0 / num_ellipses
            ellipse_width = width * (1 - opacity)
            ellipse_height = height * (1 - opacity)
            ellipse = Ellipse(
                width=ellipse_width,
                height=ellipse_height,
                color=color,
                fill_opacity=opacity,
                stroke_width=2.0,
            )
            ellipse.move_to(mean)
            ellipse.rotate(rotation)
            ellipses.add(ellipse)

        return ellipses

    def construct_simple_gaussian_ellipse(self, mean, covariance, color=ORANGE):
        """Returns a 2d Gaussian distribution object with given mean and covariance"""
        # Map mean and covariance to frame coordinates
        mean = self.axes.coords_to_point(*mean)
        angle, width, height = self.compute_covariance_rotation_and_scale(covariance)
        # Make covariance ellipses
        ellipses = VGroup()
        opacity = 0.4
        ellipse = Ellipse(
            width=width,
            height=height,
            color=color,
            fill_opacity=opacity,
            stroke_width=1.0,
        )
        ellipse.move_to(mean)
        ellipse.rotate(angle)
        ellipses.add(ellipse)
        ellipses.set_z_index(3)

        return ellipses



from manim import *
from manim_ml.neural_network.layers.feed_forward import FeedForwardLayer
from manim_ml.neural_network.layers.parent_layers import ConnectiveLayer
from manim_ml.neural_network.layers.vector import VectorLayer


class FeedForwardToVector(ConnectiveLayer):
    """Image Layer to FeedForward layer"""

    input_class = FeedForwardLayer
    output_class = VectorLayer

    def __init__(
        self,
        input_layer,
        output_layer,
        animation_dot_color=RED,
        dot_radius=0.05,
        **kwargs
    ):
        super().__init__(input_layer, output_layer, **kwargs)
        self.animation_dot_color = animation_dot_color
        self.dot_radius = dot_radius

        self.feed_forward_layer = input_layer
        self.vector_layer = output_layer

    def construct_layer(
        self,
        input_layer: "NeuralNetworkLayer",
        output_layer: "NeuralNetworkLayer",
        **kwargs
    ):
        return super().construct_layer(input_layer, output_layer, **kwargs)

    def make_forward_pass_animation(self, layer_args={}, **kwargs):
        """Makes dots diverge from the given location and move to the feed forward nodes decoder"""
        animations = []
        # Move the dots to the centers of each of the nodes in the FeedForwardLayer
        destination = self.vector_layer.get_center()
        for node in self.feed_forward_layer.node_group:
            new_dot = Dot(
                node.get_center(),
                radius=self.dot_radius,
                color=self.animation_dot_color,
            )
            per_node_succession = Succession(
                Create(new_dot),
                new_dot.animate.move_to(destination),
            )
            animations.append(per_node_succession)

        animation_group = AnimationGroup(*animations)
        return animation_group

    @override_animation(Create)
    def _create_override(self):
        return AnimationGroup()



from manim import *
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from PIL import Image
import io

def convert_matplotlib_figure_to_image_mobject(fig, dpi=200):
    """Takes a matplotlib figure and makes an image mobject from it

    Parameters
    ----------
    fig : matplotlib figure
        matplotlib figure
    """
    fig.tight_layout(pad=0)
    # plt.axis('off')
    fig.canvas.draw()
    # Save data into a buffer
    image_buffer = io.BytesIO()
    plt.savefig(image_buffer, format='png', dpi=dpi)
    # Reopen in PIL and convert to numpy
    image = Image.open(image_buffer)
    image = np.array(image)
    # Convert it to an image mobject
    image_mobject = ImageMobject(image, image_mode="RGB")

    return image_mobject


import numpy as np
from manim import *

from manim_ml.neural_network.layers.convolutional_2d_to_convolutional_2d import (
    Convolutional2DToConvolutional2D,
    Filters,
)
from manim_ml.neural_network.layers.max_pooling_2d import MaxPooling2DLayer
from manim_ml.neural_network.layers.parent_layers import ConnectiveLayer, ThreeDLayer
from manim_ml.neural_network.layers.feed_forward import FeedForwardLayer
from manim_ml.neural_network.layers.convolutional_2d import Convolutional2DLayer

from manim.utils.space_ops import rotation_matrix


class MaxPooling2DToConvolutional2D(Convolutional2DToConvolutional2D):
    """Feed Forward to Embedding Layer"""

    input_class = MaxPooling2DLayer
    output_class = Convolutional2DLayer

    def __init__(
        self,
        input_layer: MaxPooling2DLayer,
        output_layer: Convolutional2DLayer,
        passing_flash_color=ORANGE,
        cell_width=1.0,
        stroke_width=2.0,
        show_grid_lines=False,
        **kwargs
    ):
        input_layer.num_feature_maps = output_layer.num_feature_maps
        super().__init__(input_layer, output_layer, **kwargs)
        self.passing_flash_color = passing_flash_color
        self.cell_width = cell_width
        self.stroke_width = stroke_width
        self.show_grid_lines = show_grid_lines

    def construct_layer(
        self,
        input_layer: "NeuralNetworkLayer",
        output_layer: "NeuralNetworkLayer",
        **kwargs
    ):
        """Constructs the MaxPooling to Convolution3D layer

        Parameters
        ----------
        input_layer : NeuralNetworkLayer
            input layer
        output_layer : NeuralNetworkLayer
            output layer
        """
        super().construct_layer(input_layer, output_layer, **kwargs)



from manim import *
from abc import ABC, abstractmethod

class NeuralNetworkLayer(ABC, Group):
    """Abstract Neural Network Layer class"""

    def __init__(self, text=None, *args, **kwargs):
        super(Group, self).__init__()
        self.title_text = kwargs["title"] if "title" in kwargs else " "
        self.title = Text(self.title_text, font_size=DEFAULT_FONT_SIZE / 3).scale(0.6)
        self.title.next_to(self, UP, 1.2)
        # self.add(self.title)

    @abstractmethod
    def construct_layer(
        self,
        input_layer: "NeuralNetworkLayer",
        output_layer: "NeuralNetworkLayer",
        **kwargs,
    ):
        """Constructs the layer at network construction time

        Parameters
        ----------
        input_layer : NeuralNetworkLayer
            preceding layer
        output_layer : NeuralNetworkLayer
            following layer
        """
        if "debug_mode" in kwargs and kwargs["debug_mode"]:
            self.add(SurroundingRectangle(self))

    @abstractmethod
    def make_forward_pass_animation(self, layer_args={}, **kwargs):
        pass

    @override_animation(Create)
    def _create_override(self):
        return Succession()

    def __repr__(self):
        return f"{type(self).__name__}"

class VGroupNeuralNetworkLayer(NeuralNetworkLayer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # self.camera = camera

    @abstractmethod
    def make_forward_pass_animation(self, **kwargs):
        pass

    @override_animation(Create)
    def _create_override(self):
        return super()._create_override()

class ThreeDLayer(ABC):
    """Abstract class for 3D layers"""
    pass
    # Angle of ThreeD layers is static context


class ConnectiveLayer(VGroupNeuralNetworkLayer):
    """Forward pass animation for a given pair of layers"""

    @abstractmethod
    def __init__(self, input_layer, output_layer, **kwargs):
        super(VGroupNeuralNetworkLayer, self).__init__(**kwargs)
        self.input_layer = input_layer
        self.output_layer = output_layer
        # Handle input and output class
        # assert isinstance(input_layer, self.input_class), f"{input_layer}, {self.input_class}"
        # assert isinstance(output_layer, self.output_class), f"{output_layer}, {self.output_class}"

    @abstractmethod
    def make_forward_pass_animation(self, run_time=2.0, layer_args={}, **kwargs):
        pass

    @override_animation(Create)
    def _create_override(self):
        return super()._create_override()

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            + f"input_layer={self.input_layer.__class__.__name__},"
            + f"output_layer={self.output_layer.__class__.__name__},"
            + ")"
        )


class BlankConnective(ConnectiveLayer):
    """Connective layer to be used when the given pair of layers is undefined"""

    def __init__(self, input_layer, output_layer, **kwargs):
        super().__init__(input_layer, output_layer, **kwargs)

    def make_forward_pass_animation(self, run_time=1.5, layer_args={}, **kwargs):
        return AnimationGroup(run_time=run_time)

    @override_animation(Create)
    def _create_override(self):
        return super()._create_override()



from manim import *
from manim_ml.utils.mobjects.probability import GaussianDistribution
from manim_ml.neural_network.layers.parent_layers import VGroupNeuralNetworkLayer


class EmbeddingLayer(VGroupNeuralNetworkLayer):
    """NeuralNetwork embedding object that can show probability distributions"""

    def __init__(
        self,
        point_radius=0.02,
        mean=np.array([0, 0]),
        covariance=np.array([[1.0, 0], [0, 1.0]]),
        dist_theme="gaussian",
        paired_query_mode=False,
        **kwargs
    ):
        super(VGroupNeuralNetworkLayer, self).__init__(**kwargs)
        self.mean = mean
        self.covariance = covariance
        self.gaussian_distributions = VGroup()
        self.add(self.gaussian_distributions)
        self.point_radius = point_radius
        self.dist_theme = dist_theme
        self.paired_query_mode = paired_query_mode

    def construct_layer(
        self,
        input_layer: "NeuralNetworkLayer",
        output_layer: "NeuralNetworkLayer",
        **kwargs
    ):
        self.axes = Axes(
            tips=False,
            x_length=0.8,
            y_length=0.8,
            x_range=(-1.4, 1.4),
            y_range=(-1.8, 1.8),
            x_axis_config={"include_ticks": False, "stroke_width": 0.0},
            y_axis_config={"include_ticks": False, "stroke_width": 0.0},
        )
        self.add(self.axes)
        self.axes.move_to(self.get_center())
        # Make point cloud
        self.point_cloud = self.construct_gaussian_point_cloud(
            self.mean, self.covariance
        )
        self.add(self.point_cloud)
        # Make latent distribution
        self.latent_distribution = GaussianDistribution(
            self.axes, mean=self.mean, cov=self.covariance
        )  # Use defaults
        super().construct_layer(input_layer, output_layer, **kwargs)

    def add_gaussian_distribution(self, gaussian_distribution):
        """Adds given GaussianDistribution to the list"""
        self.gaussian_distributions.add(gaussian_distribution)

        return Create(gaussian_distribution)

    def remove_gaussian_distribution(self, gaussian_distribution):
        """Removes the given gaussian distribution from the embedding"""
        for gaussian in self.gaussian_distributions:
            if gaussian == gaussian_distribution:
                self.gaussian_distributions.remove(gaussian_distribution)
                return FadeOut(gaussian)

    def sample_point_location_from_distribution(self):
        """Samples from the current latent distribution"""
        mean = self.latent_distribution.mean
        cov = self.latent_distribution.cov
        point = np.random.multivariate_normal(mean, cov)
        # Make dot at correct location
        location = self.axes.coords_to_point(point[0], point[1])

        return location

    def get_distribution_location(self):
        """Returns mean of latent distribution in axes frame"""
        return self.axes.coords_to_point(self.latent_distribution.mean)

    def construct_gaussian_point_cloud(
        self, mean, covariance, point_color=WHITE, num_points=400
    ):
        """Plots points sampled from a Gaussian with the given mean and covariance"""
        # Sample points from a Gaussian
        np.random.seed(5)
        points = np.random.multivariate_normal(mean, covariance, num_points)
        # Add each point to the axes
        point_dots = VGroup()
        for point in points:
            point_location = self.axes.coords_to_point(*point)
            dot = Dot(point_location, color=point_color, radius=self.point_radius / 2)
            dot.set_z_index(-1)
            point_dots.add(dot)

        return point_dots

    def make_forward_pass_animation(self, layer_args={}, **kwargs):
        """Forward pass animation"""
        animations = []
        if "triplet_args" in layer_args:
            triplet_args = layer_args["triplet_args"]
            positive_dist_args = triplet_args["positive_dist"]
            negative_dist_args = triplet_args["negative_dist"]
            anchor_dist_args = triplet_args["anchor_dist"]
            # Create each dist
            anchor_dist = GaussianDistribution(self.axes, **anchor_dist_args)
            animations.append(Create(anchor_dist))

            positive_dist = GaussianDistribution(self.axes, **positive_dist_args)
            animations.append(Create(positive_dist))

            negative_dist = GaussianDistribution(self.axes, **negative_dist_args)
            animations.append(Create(negative_dist))
            # Draw edges in between anchor and positive, anchor and negative
            anchor_positive = Line(
                anchor_dist.get_center(),
                positive_dist.get_center(),
                color=GOLD,
                stroke_width=DEFAULT_STROKE_WIDTH / 2,
            )
            anchor_positive.set_z_index(3)
            animations.append(Create(anchor_positive))

            anchor_negative = Line(
                anchor_dist.get_center(),
                negative_dist.get_center(),
                color=GOLD,
                stroke_width=DEFAULT_STROKE_WIDTH / 2,
            )
            anchor_negative.set_z_index(3)

            animations.append(Create(anchor_negative))
        elif not self.paired_query_mode:
            # Normal embedding mode
            if "dist_args" in layer_args:
                scale_factor = 1.0
                if "scale_factor" in layer_args:
                    scale_factor = layer_args["scale_factor"]
                self.latent_distribution = GaussianDistribution(
                    self.axes, **layer_args["dist_args"]
                ).scale(scale_factor)
            else:
                # Make ellipse object corresponding to the latent distribution
                # self.latent_distribution = GaussianDistribution(
                #     self.axes,
                #     dist_theme=self.dist_theme,
                #     cov=np.array([[0.8, 0], [0.0, 0.8]])
                # )
                pass
            # Create animation
            create_distribution = Create(self.latent_distribution)
            animations.append(create_distribution)
        else:
            # Paired Query Mode
            assert "positive_dist_args" in layer_args
            assert "negative_dist_args" in layer_args
            positive_dist_args = layer_args["positive_dist_args"]
            negative_dist_args = layer_args["negative_dist_args"]
            # Handle logic for embedding a paired query into the embedding layer
            positive_dist = GaussianDistribution(self.axes, **positive_dist_args)
            self.gaussian_distributions.add(positive_dist)
            negative_dist = GaussianDistribution(self.axes, **negative_dist_args)
            self.gaussian_distributions.add(negative_dist)

            animations.append(Create(positive_dist))
            animations.append(Create(negative_dist))

        animation_group = AnimationGroup(*animations)

        return animation_group

    @override_animation(Create)
    def _create_override(self, **kwargs):
        # Plot each point at once
        point_animations = []
        for point in self.point_cloud:
            point_animations.append(GrowFromCenter(point))

        point_animation = AnimationGroup(*point_animations, lag_ratio=1.0, run_time=2.5)

        return point_animation


class NeuralNetworkEmbeddingTestScene(Scene):
    def construct(self):
        nne = EmbeddingLayer()
        mean = np.array([0, 0])
        cov = np.array([[5.0, 1.0], [0.0, 1.0]])

        point_cloud = nne.construct_gaussian_point_cloud(mean, cov)
        nne.add(point_cloud)

        gaussian = nne.construct_gaussian_distribution(mean, cov)
        nne.add(gaussian)

        self.add(nne)



from manim import *
from manim_ml.neural_network.layers.convolutional_2d_to_feed_forward import (
    Convolutional2DToFeedForward,
)
from manim_ml.neural_network.layers.feed_forward import FeedForwardLayer
from manim_ml.neural_network.layers.max_pooling_2d import MaxPooling2DLayer


class MaxPooling2DToFeedForward(Convolutional2DToFeedForward):
    """Feed Forward to Embedding Layer"""

    input_class = MaxPooling2DLayer
    output_class = FeedForwardLayer

    def __init__(
        self,
        input_layer: MaxPooling2DLayer,
        output_layer: FeedForwardLayer,
        passing_flash_color=ORANGE,
        **kwargs
    ):
        super().__init__(input_layer, output_layer, **kwargs)

    def construct_layer(
        self,
        input_layer: "NeuralNetworkLayer",
        output_layer: "NeuralNetworkLayer",
        **kwargs
    ):
        return super().construct_layer(input_layer, output_layer, **kwargs)



from manim import *
from PIL import Image

from manim_ml.neural_network.layers.convolutional_2d import Convolutional2DLayer
from manim_ml.neural_network.layers.feed_forward import FeedForwardLayer
from manim_ml.neural_network.layers.image import ImageLayer
from manim_ml.neural_network.neural_network import NeuralNetwork


class SingleConvolutionalLayerScene(ThreeDScene):
    def construct(self):
        # Make nn
        layers = [Convolutional2DLayer(3, 4)]
        nn = NeuralNetwork(layers)
        nn.scale(1.3)
        # Center the nn
        nn.move_to(ORIGIN)
        self.add(nn)
        # Play animation
        self.set_camera_orientation(
            phi=280 * DEGREES, theta=-10 * DEGREES, gamma=90 * DEGREES
        )
        # self.play(nn.make_forward_pass_animation(run_time=5))


class Simple3DConvScene(ThreeDScene):
    def construct(self):
        """
        TODO
        - [X] Make grid lines for the CNN filters
        - [ ] Make Scanning filter effect
        - [ ] Have filter box go accross each input feature map
        - [ ] Make filter lines effect
        - [ ] Make flowing animation down filter lines
        """
        # Make nn
        layers = [
            Convolutional2DLayer(num_feature_maps=1, feature_map_size=3, filter_size=3),
            Convolutional2DLayer(num_feature_maps=1, feature_map_size=3, filter_size=3),
        ]
        nn = NeuralNetwork(layers)
        # Center the nn
        nn.move_to(ORIGIN)
        self.add(nn)
        # Play animation
        # self.set_camera_orientation(phi=280*DEGREES, theta=-10*DEGREES, gamma=90*DEGREES)
        self.play(nn.make_forward_pass_animation(run_time=30))


# Make the specific scene
config.pixel_height = 1200
config.pixel_width = 1900
config.frame_height = 6.0
config.frame_width = 6.0


class CombinedScene(ThreeDScene):
    def construct(self):
        image = Image.open("../assets/mnist/digit.jpeg")
        numpy_image = np.asarray(image)
        # Make nn
        nn = NeuralNetwork(
            [
                ImageLayer(numpy_image, height=1.5),
                Convolutional2DLayer(1, 7, filter_spacing=0.32),
                Convolutional2DLayer(3, 5, 3, filter_spacing=0.32),
                Convolutional2DLayer(5, 3, 3, filter_spacing=0.18),
                FeedForwardLayer(3),
                FeedForwardLayer(3),
            ],
            layer_spacing=0.25,
        )
        # Center the nn
        nn.move_to(ORIGIN)
        self.add(nn)
        # Play animation
        forward_pass = nn.make_forward_pass_animation()
        self.wait(1)
        self.play(forward_pass)
